/*
Original by Julien
Heavily modified 2008 by Sven Killig <sven@killig.de>
*/

#ifndef _4D_H
#define _4D_H

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <asm/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <pthread.h>
#include <termios.h>
#include <string>
#include <vector>
#include <iostream>
#include <poll.h>

//for 4 bytes long variables
#define address_mmsb(variable) ((variable >> 24) & 0xFF)
#define address_mlsb(variable) ((variable >> 16) & 0xFF)
#define address_lmsb(variable) ((variable >> 8) & 0xFF)
#define address_llsb(variable) (variable & 0xFF)

//for 3 bytes long variables
#define address_hi(variable) (variable >> 16)
#define address_mid(variable) ((variable >> 8) & 0xFF)
#define address_lo(variable) (variable & 0xFF)

//for 2 bytes long variables
#define msb(variable) (variable >> 8)
#define lsb(variable) (variable & 0xFF)

//#define debug	1
#define USE_THREAD TRUE

//Copied from linux/fb.h

struct var_screeninfo {
	std::string id;			/* identification string eg "TT Builtin" */
	__u32 xres;			/* visible resolution		*/
	__u32 yres;
	__u32 xres_virtual;		/* virtual resolution		*/
	__u32 yres_virtual;
	__u32 xoffset;			/* offset from virtual to visible */
	__u32 yoffset;			/* resolution			*/
	__u32 bits_per_pixel;		/* guess what			*/
	__u32 height;			/* height of picture in mm    */
	__u32 width;			/* width of picture in mm     */
	__u32 baud_rate;		/*added */
	__u32 timeout;			/*time to wait for ACK*/
	bool writeWaitForACK;		/*added, 1 = Wait ACK */
	const char * deviceName;	/*added, name of the communication port */
	__u32 hardwareVersion;
	__u32 firmwareVersion;
};

//generic class to manage 4D modules
class _4DModule
{
	public:
		_4DModule(void);					//constructor
		_4DModule(const char * deviceName);
		_4DModule(const char * deviceName, int baudrate);
		//Generic methods
		bool modup(void);					//do all the necessary step to bring up the module
		bool moddown(void);					//do all the necessary step to bring down the module
		bool modconfig(const char * deviceName);
		bool modconfig(int baudrate);
		bool modconfig(const char * deviceName, int baudrate);
		bool isOpen(void);
		void setTimeout(unsigned int time);			//Time is in ms;
		int  initialise(void);					//initialise the module
		int  write(const char * buffer, int bytes, char pos1, char pos2);
		int  write(const char * buffer, int bytes);		//direct access to the module to write <bytes> bytes
		int  write(unsigned int * buffer, int bytes);
		int  writeWaitForACK(const char * buffer, int bytes);
		int  writeWaitForACK(unsigned int * buffer, int bytes);
		int read(const char * buffer, int bytes);		//direct access to the module to read <bytes> bytes
		int readTimeout(const char * buffer, int bytes);
		void writeWaitForACK(bool mode);			//does write wait for ACK from module ? 1 = yes
		var_screeninfo getScreenInfo(void);			//returns infos like screen width, height etc... see struct below
		bool displayScreenInfo(void);
		bool reset(void);					//reset the module

		//General command set:
		bool addCharacter(unsigned char number,
				std::vector<unsigned char> * data);

		bool setBackground(short int color);			//all modules

		bool placeButton(bool state, char x, short int y,	//all modules except uLCD xxx (0Mb)
				short int color, char font,
				short int textColor,
				char textWidth,
				char textHeight,
				std::string * buttonText);
		bool placeButton(bool state, char x, short int y,	//uLCD xxx PMD2
				short int color, char font,
				short int textColor,
				char textWidth,
				char textHeight,
				const char * buttonText);
		bool placeButton(bool state, char x, char y,	//all modules except uLCD xxx (0Mb)
				short int color, char font,
				short int textColor,
				char textWidth,
				char textHeight,
				std::string * buttonText);
		bool placeButton(bool state, char x, char y,	//uLCD xxx PMD2
				short int color, char font,
				short int textColor,
				char textWidth,
				char textHeight,
				const char * buttonText);
		bool placeButton(bool state, char x1, char y1,		//for uOLED xxx
				char x2, char y2, short int color);

		bool drawCircle(char x, short int y, char radius,	//all modules
				short int color);
		bool drawCircle(char x, char y, char radius,	//all modules
				short int color);
		bool drawCircleFill(char x, char y, char radius,  //uOLED xxx, uLCD xxx xMb
				short int color);

		bool drawEllipse(char x, short int y, char rx,		//uLCD xxx PMD2
				short int ry, short int color);

		bool copyPaste(char x, short int y, char xd,		//all modules
				short int yd, char width,
				short int height);
// 		bool copyPaste(char xs, char ys, char xd, char yd,
// 				char width, char height);
		bool copyPaste(char x1, char y1, char x2, char y2,
				char x3, char y3);
		bool copyPaste(short int xs, short int ys, short int xd,
				short int yd, short int width,
				short int height, char sPage, char dPage);

		bool displayCharacter(char group, char number, char x,	//all modules
				short int y, short int color);
		bool displayCharacter(char number, char x, char y,
				short int color);

		bool eraseScreen(void);					//all modules
		bool setFontSize(char size);				//all modules

		bool drawTriangle(char x1, short int y1,		//uOLED 1xx GMD1, uLCD xxx PMD2
				char x2, short int y2,
				char x3, short int y3,
				short int color);
		bool drawTriangle(char x1, char y1,
				char x2, char y2,
				char x3, char y3,
				short int color);

		bool drawPolygon(std::vector<short int> * vertices,//uOLED 1xx GMD1, uLCD xxx PMD2
				unsigned int color);
		bool drawPolygon(std::vector<char> * vertices,//uOLED 1xx GMD1, uLCD xxx PMD2
				unsigned int color);

		bool displayImage(unsigned char x, unsigned short int y, unsigned char width, unsigned short int height, char colorMode, std::vector<unsigned short int> * pixels);

		bool drawLine(char x1, short int y1,			//all modules
				char x2, short int y2,
				short int color);

		bool setTextOpacity(bool mode);				//all modules

		bool putPixel(char x, short int y, short int color);	//all modules
		bool putPixel(char x, char y, short int color);

		bool setPenSize(char size);				//uOLED 1xx GMD1, uLCD xxx PMD2

		int  readPixel(char x, short int y);			//all modules
		int  readPixel(char x, char y);

		bool drawRectangle(unsigned char x1, short int y1, unsigned char x2, short int y2, short int color);

		bool paintArea(char x1, char y1, char x2, char y2,	//uOLED xxx, uLCD xxx (0Mb), uVGA core
				short int color);

		bool placeString(char x, short int y, char font,	//formatted: all modules except uLCD xxx (0Mb)
				short int color, char width,
				char height,
				std::string * character);
		bool placeString(char x, char y, char font,
				short int color, char width,
				char height,
				std::string * character);
		bool placeString(char column, char row, char font,	//unformatted: uOLED 1xx GMD1, uLCD xxx PMD2
				short int color,
				std::string * character);
		bool placeString(char x, short int y, char font,	//formatted: all modules except uLCD xxx (0Mb)
				short int color, char width,
				char height,
				const char * character);
		bool placeString(char x, char y, char font,
				short int color, char width,
				char height,
				const char * character);
		bool placeString(char column, char row, char font,	//unformatted: uOLED 1xx GMD1, uLCD xxx PMD2
				short int color,
				const char * character);

		bool placeText(char character, char column, char row,	//formatted: all modules
				short int color);
		bool placeText(char character, char x, short int y,	//unformatted: all modules except uLCD xxx (0Mb)
				short int color, char width,
				char height);
		bool placeText(char character, char x, char y,
				short int color, char width,
				char height);
					//uVGA core
		bool plotGraph(short int x, short int y, char color,
				short int size, char delay, bool erase,
				char eraseColor, char eraseHeight,
				std::vector<char> * data);

		bool setBacklight(char mode);

		bool setDisplay(bool mode);

		bool setContrast(char value);

		bool powerUp(void);
		bool powerDown(void);

		//Display specific command set:
		bool writeRegister(char reg, char data);			//uOLED 1xx GMD1
		bool writeRegister(char data, bool mode);

		bool displayScroll(bool enable, bool direction,	//uOLED 1xx GMD1
				char speed);
		bool displayScroll(bool enable, char speed);

		bool dimScreen(char x, char y, char width, char height);//uOLED 128 GMD1

		bool setFloatPointerParams(bool mode, bool appearance,	//uLCD xxx PMD2
					  bool size, short int color);
		bool moveFloatPointer(char x, short int y);		//uLCD xxx PMD2
		std::vector<int>  getFloatPointerCoord(void);		//uLCD xxx PMD2
		bool vsyncLock(void);					//uLCD xxx PMD2, uVGA core

		//Extended command set:					//All uSD specific to GMD1 module and forth
		bool uSDInit(void);

		std::vector<unsigned char> uSDReadSector(int address);

		bool uSDWriteSector(int address, std::vector<unsigned char> * data);

		unsigned char uSDReadByte(void);
		bool uSDWriteByte(char byte);
		bool uSDSetAddress(int address);

		bool uSDCopyScreen(char x, char y, char width, char height, int address);
		bool uSDDisplayImage(char x, char y, char width, char height, char colourMode, int address);

		bool uSDDisplayObject(int address);
		bool uSDRun(int address);
		bool uSDExit(void);
		/*bool fmDisplayImage(int address);			//only for flash enabled modules
		bool fmDisplayObject(char x, char y,
				    int address);
		bool fmRun(void);
		bool fmExit(void);
		bool fmWrite(short int pageNumber,
				std::vector<char> * data);
		int  fmRead(short int pageNumber);
		bool fmErase(void);
		int  vmRead(char page);					//only for VGA modules
		bool vmWrite(char page);
		bool vmDisplay(char page);*/
	private:
		//serial communication handler
		bool  open(void);
		void  close(void);
		bool  setBaudRate(int baudrate);
		bool __getACK(void);					//private method to get ACK from the module
		bool __getScreenInfo(void);

		int portHandle;
		var_screeninfo screeninfo;
};

                int RGB_to16(int red, int grn, int blu);

#endif
